import {readdir, stat, readFile} from 'fs/promises';
import {resolve} from 'path';
import {glob} from 'glob';
import * as v from 'valibot';
import {
	Binding,
	BINDING_FILENAME,
	Book,
	BOOK_FILENAME, BookData,
	bookFileSchema,
	BookManifestItem,
	CommonInput,
} from '../../common';
import { randomUUID } from 'crypto';

export const name = 'dir' as const;

export const inputSchema = v.object({
	sourceType: v.literal(name),
	path: v.string(),
});

interface DirInput extends CommonInput, v.Output<typeof inputSchema> {}

export class InvalidInputPathError extends Error {}

const getBookFile = async (bookFilePath: string, packageJsonFilePath: string): Promise<Book> => {
	const bookFileString = await readFile(bookFilePath, 'utf-8');
	const bookFileRaw = JSON.parse(bookFileString);

	const packageJsonFileString = await readFile(packageJsonFilePath, 'utf-8');
	const packageJson = JSON.parse(packageJsonFileString);

	const bookFile = {
		...bookFileRaw,
		id: bookFileRaw.id ?? randomUUID(),
		publisher: bookFileRaw.publisher ?? '@modal-sh/patchouli',
		description: bookFileRaw.description ?? packageJson.description,
		rights: bookFileRaw.rights ?? packageJson.copyright,
	};

	return await v.parseAsync(bookFileSchema, bookFile);
};

const getBindingFile = async (bindingFilePath: string, defaultBinding: Binding): Promise<Binding> => {
	const bindingFileString = await readFile(bindingFilePath, 'utf-8');
	const bindingFileRaw = JSON.parse(bindingFileString);

	return {
		...defaultBinding,
		...bindingFileRaw,
	};
};

const getItems = async (dir: string, prefix = ''): Promise<BookManifestItem[]> => {
	const thisDirFiles = await readdir(dir);

	const items = await Promise.all(
		thisDirFiles.map(async (p) => {
			const thisPath = resolve(dir, p);
			const s = await stat(thisPath);
			if (s.isDirectory()) {
				return getItems(thisPath, p);
			}

			const href = [prefix, p].filter((s) => s.trim()).join('/');
			const [idRaw] = href.split('.');
			const id = idRaw.replace(/\//g, '-');

			return [
				{
					id,
					href,
					path: thisPath,
				},
			];
		})
	);

	return items.flat();
};

const getSpine = async (binding: Binding) => {
	const {
		pageOrdering,
		generatorDistDirectory,
		pageFileExtensions,
	} = binding;

	if (!Array.isArray(pageOrdering)) {
		// get glob of HTML pages
		return [];
	}

	const directoryNormalized = generatorDistDirectory.replace(/\\/g, '/');
	const extensionsGlobFragment = pageFileExtensions.join(',');

	const pagesPaths = await Promise.all(
		pageOrdering.map(
			async (globString) => {
				switch (globString) {
					case 'index':
					case 'title':
						return glob(`${directoryNormalized}/${globString}{${extensionsGlobFragment}`);
					case 'toc':
						return []; // TODO check if TOC is pre-generated or there is a directive to generate a TOC
					default:
						break;
				}

				if (globString.startsWith('pages:')) {
					const [prefix, ...etcGlobStrArr] = globString.split(':');
					const globStrPath = etcGlobStrArr.join(':');
					switch (prefix) {
						case 'pages': {
							const thePagesPath = await glob(`${directoryNormalized}/${globStrPath}{${extensionsGlobFragment}}`);
							return thePagesPath.sort((a, b) => a.localeCompare(b));
						}
					}
				}

				return [];
			}
		)
	);

	return pagesPaths.flat();
};

export const compileFromInput = async <T extends CommonInput = DirInput>(inputRaw: T): Promise<BookData> => {
	const input = inputRaw as unknown as DirInput;
	const files = await readdir(input.path);
	if (!files.includes(BOOK_FILENAME)) {
		throw new InvalidInputPathError(`Path does not contain a "${BOOK_FILENAME}" file.`);
	}

	const bookFilePath = resolve(input.path, BOOK_FILENAME);
	const packageJsonFilePath = resolve(input.path, 'package.json');
	const bookFile = await getBookFile(bookFilePath, packageJsonFilePath);

	const defaultBinding: Binding = {
		generatorType: 'static',
		generatorDistDirectory: resolve(input.path, 'dist'),
		pageFileExtensions: ['.html', '.htm', '.xhtml'],
	};

	const bindingFilePath = resolve(input.path, BINDING_FILENAME);
	const bindingFile: Binding = files.includes(BINDING_FILENAME)
		? await getBindingFile(bindingFilePath, defaultBinding)
		: defaultBinding;

	const isAstro = files.includes('astro.config.mjs');
	if (isAstro) {
		bindingFile.generatorType = 'astro';
		bindingFile.generatorConfigFilePath = resolve(input.path, 'astro.config.mjs');
		const { default: config } = await import('file:///' + bindingFile.generatorConfigFilePath);
		bindingFile.generatorDistDirectory = resolve(input.path, config.outDir ?? 'dist');
		if (typeof bindingFile.pageOrdering !== 'object') {
			bindingFile.pageOrdering = []; // if we want to have a custom ordering, e.g. last minute changes
		}
	}

	const spine = await getSpine(bindingFile);
	const manifest = await getItems(
		bindingFile.generatorDistDirectory
	);

	return {
		metadata: bookFile,
		manifest,
		spine,
		binding: bindingFile,
	};
};
