import {Readable} from 'stream';
import {BookData} from '../../common';
import {prepareEpubContainerContents, prepareEpubRootfileContents} from './xml-definitions';
import {createReadStream} from 'fs';
import assert from 'assert';
import archiver from 'archiver';

interface EpubArchiveDatum {
	path: string;
	data: Buffer | Readable;
}

export const prepareEpubContents = async <U extends BookData = BookData>(bookPackage: U): Promise<EpubArchiveDatum[]> => {
	const baseDir = 'EPUB';
	const rootFilePath = `${baseDir}/content.opf`;
	const rootFileContents = await prepareEpubRootfileContents(bookPackage);
	return [
		{
			path: 'mimetype',
			data: Buffer.from('application/epub+zip'),
		},
		{
			path: 'META-INF/container.xml',
			data: Buffer.from(prepareEpubContainerContents(rootFilePath))
		},
		{
			path: rootFilePath,
			data: Buffer.from(rootFileContents),
		},
		...bookPackage.manifest.map((entry) => {
			if (typeof entry.path === 'string') {
				const finalPath = entry.path.slice(bookPackage.binding.generatorDistDirectory.length ?? 0);
				return {
					path: `${baseDir}/${finalPath}`,
					data: createReadStream(entry.path),
				};
			}

			assert(entry.buffer instanceof Buffer);
			return {
				path: `${baseDir}/${entry.href}`,
				data: entry.buffer,
			};
		}),
	];
};

export const createEpubArchive = async (data: EpubArchiveDatum[]): Promise<Buffer> => {
	const archive = archiver('zip', {
		zlib: {
			level: 9,
		},
	});

	return new Promise<Buffer>(async (resolve, reject) => {
		let b = Buffer.from('');

		archive.on('data', (c) => {
			b = Buffer.concat([b, c]);
		});

		archive.on('end', () => {
			resolve(b);
		});

		archive.on('error', (err) => {
			reject(err);
		});

		data.forEach((entry) => {
			archive.append(entry.data, {
				name: entry.path,
			});
		});

		await archive.finalize();
	});
};
